import logging
import os
import subprocess
import shlex
import threading
import uuid

from pathlib import Path
from werkzeug.utils import secure_filename

from .constants import *
from .exceptions import *


log = logging.getLogger('flask-backend')

updateStatus = None

p = Path(FW_UPDATES_DIR)
if not p.exists():
    try:
        p.mkdir()
    except Exception as e:
        log.error("Error creating directory {}: {}".format(FW_UPDATES_DIR, e))

def updateFirmware(fileUploaded):
    global updateStatus
    log.info("Uploading new fw image...")
    try:
        updateStatus = "uploading"
        if 'fwImage' not in fileUploaded:
            log.error('fwImage is not in request files uploaded')
            # remove file if error occurs
            removeUpdateFileChecker()
            raise InvalidUploadedFileError('Invalid file')
        uploadedFile = fileUploaded['fwImage']
        # if user does not select file, browser also
        # submit an empty part without filename
        if uploadedFile.filename == '':
            log.error('filename from uploaded file is invalid')
            # remove file if error occurs
            removeUpdateFileChecker()
            raise InvalidUploadedFileError('Invalid filename')
        # generate uuid and store it in file
        # if uuid exists, stop update request because there's already an update running
        # file is automatically removed after reboot
        if not os.path.isfile(FW_UPDATE_FILE_PATH):
            uid = uuid.uuid4().hex
            try:
                f = open(FW_UPDATE_FILE_PATH, 'w')
                f.write(uid)
                f.close()
            except Exception as e:
                log.error("Error while writing in file {}: {}".format(FW_UPDATE_FILE_PATH, e))
                raise e
        else:
            log.error("ABORTING UPDATE! There is an update already running.")
            raise FwUpdateRunningError()
        filename = secure_filename(uploadedFile.filename)
        try:
            log.debug("Storing fw update file in {}".format(FW_UPDATES_DIR))
            uploadedFile.save(os.path.join(FW_UPDATES_DIR, filename))
        except Exception as e:
            log.error(e)
            updateStatus = "error"
            # remove file if error occurs
            removeUpdateFileChecker()
            raise e
        t = threading.Thread(target=installNewFW,
                             args=(os.path.join(FW_UPDATES_DIR, secure_filename(uploadedFile.filename)),))
        t.start()
    except Exception as e:
        log.error("Error occurred while uploading fw file : {}".format(e))
        updateStatus = "error"
        # remove file if error occurs
        removeUpdateFileChecker()
        if str(e).startswith('413 Request Entity Too Large'):
            raise FileTooLargeError('REQUEST ENTITY TOO LARGE')
        else:
            raise e

def installNewFW(filePath):
    global updateStatus
    updateStatus = "installing"
    cmd = "qiba-update-client -f "+filePath
    log.info("Executing: {}".format(cmd))
    proc = None
    try:
        proc = subprocess.Popen(shlex.split(cmd), shell=False)
        proc.wait()
    except Exception as e:
        log.error(e)
        updateStatus = "error"
        # remove file if error occurs
        removeUpdateFileChecker()
    #log.debug("qiba-update-client returned {}".format(proc.returncode))
    if proc.returncode != 0:
        log.error("qiba-update-client failed installation")
        updateStatus = "error"
        removeUpdateFileChecker()
        # remove file
        try:
            os.remove(filePath)
        except Exception as e:
            log.error("Can't remove file {}: {}".format(filePath, e))
            # remove file if error occurs
            removeUpdateFileChecker()
    else:
        # remove file after installing it
        try:
            os.remove(filePath)
        except Exception as e:
            log.error("Can't remove file {}: {}".format(filePath, e))
            updateStatus = "error"
            # remove file if error occurs
            removeUpdateFileChecker()
        cmd = "reboot"
        log.info("Executing: {}".format(cmd))
        subprocess.Popen(shlex.split(cmd), shell=False)
        updateStatus = "rebooting"
        # no need to removeUpdateFileChecker because device will reboot and /tmp will be cleaned

def removeUpdateFileChecker():
    try:
        if os.path.isfile(FW_UPDATE_FILE_PATH):
            os.remove(FW_UPDATE_FILE_PATH)
    except Exception as e:
        log.error("Can't remove file {}: {}".format(FW_UPDATE_FILE_PATH, e))

def getFwUpdateStatus():
    log.debug("Firmware Update status: {}".format(updateStatus))
    return updateStatus